package anywheresoftware.b4a.agraham.dialogs;

import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.GregorianCalendar;
import java.util.NoSuchElementException;

import android.app.AlertDialog;
import android.content.DialogInterface;
import android.graphics.Bitmap;
import android.graphics.Color;
import android.graphics.drawable.BitmapDrawable;
import android.graphics.drawable.GradientDrawable;
import android.text.method.PasswordTransformationMethod;
import android.view.Gravity;
import android.view.View;
import android.view.ViewGroup;
import android.view.View.OnClickListener;
import android.view.inputmethod.InputMethodManager;
import android.widget.AbsoluteLayout;
import android.widget.AdapterView;
import android.widget.Button;
import android.widget.DatePicker;
import android.widget.EditText;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.SeekBar;
import android.widget.TableLayout;
import android.widget.TableRow;
import android.widget.TextView;
import android.widget.TimePicker;
import android.widget.LinearLayout.LayoutParams;
import android.widget.SeekBar.OnSeekBarChangeListener;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.Msgbox;
import anywheresoftware.b4a.BA.ActivityObject;
import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;
import anywheresoftware.b4a.keywords.Common;
import anywheresoftware.b4a.keywords.DateTime;
import anywheresoftware.b4a.keywords.constants.Colors;
import anywheresoftware.b4a.objects.SimpleListAdapter;

/**
 *This modal dialog allows the collection of user entered data in the form of text.
 *The default is free text but the input can be restricted to numeric characters only
 *or to signed numbers including a decimal point.
 */
@SuppressWarnings("deprecation")
@ShortName("InputDialog")
@ActivityObject
@Version(2.9f)
@Author("Andrew Graham")
public class InputDialog
{
	// 1.0 initial release
	// 1.1 setDateTicks was off by one in month
	// 1.2 added ColorDialog and ColorPickerDialog
	// 1.3 fixed ColorPickerDialog for non WVGA screens size
	// 1.4 added NumberDialog
	// 1.5 added InputDialog Hint and HintColor 
	//	   addded INPUT_TYPE_PHONE and INPUT_TYPE_NONE to InputDialog
	//	   removed SingleLine property from InputDialog as it was always a single line
	//	   added ShowSign property to NumberDialog, increased max digits to 9
	// 1.6 added FileDialog
	// 1.7 added List.setCacheColorHint to FileDialog,
	//	   added logging of NoSuchElementException FileDialog
	// 1.8 reset List scroll position on folder change
	//	   added CustomDialog
	// 1.9 made FileDialog ChosenName behave properly
	// 2.0 CustomDialog and FileDialog Response were not set on return
	// 2.1 added CustomDialog2
	// 2.2 DateDialog.SetDate wasn't subtracting 1 from the supplied value.
	//	   NumberDialog removed setTextColor from + buttons in case of non-standard device themes 
	//	   NumberDialog removed setTextColor and setBackgroundColor from number buttons 
	// 2.3 removed the supplied view in Custom and Custom2 to let it be reused
	// 2.4 added close keyboard code to Show in Input, Custom and Custom2 dialogs 
	//	   FileDialog now accepts comma separated list of filters.
	// 2.5 Added ColorDialogHSV
	//	   added FileDialog.ScrollingBackgroundColor
	// 2.6 Fixed non-initialisation of FileDialog.ftypes, added FileDialog.ShowOnlyFolders
	// 2.7 Fixed setPalette in ColorPicker Dialog
	// 2.8 added clearFocus() to TimeDialog.Show and DateDialog.Show to force keyboard input to be accepted
	// 2.9 added setSingleLine() to the edittext
	private static final double version = 2.9;

	/**
	 *This library contains several modal, that is blocking, dialogs by which the user
	 *can enter data. Presently they are an InputDialog for text, a TimeDialog for times,
	 *a DateDialog for dates, both a ColorDialog and a ColorPickerDialog for colors, a
	 *NumberDialog for numbers,a FileDialog for folders and file names and a CustomDialog.
	 *
	 *Android does not provide modal dialogs but a special mechanism to permit this exists
	 *in Basic4android. The Android Activity lifetime system makes this support complicated
	 *because Activities can be created and destroyed at will by the OS. To avoid stack
	 *runaway on the GUI thread when an Activity is destroyed the stack must be unwound to
	 *the lowest level. The Basic4android modal mechanism does this by closing any modal
	 *dialog being shown and exiting the Sub that called the dialog, and any Sub that called 
	 *that Sub and so on, in order to return the main thread to the message loop. This means
	 *that the application does not necessarily receive a return value from the dialog and has
	 *its expected flow of execution interrupted. This will probably most often happen if the
	 *device is rotated while a modal dialog is displayed so the Activity is destroyed and rebuilt
	 *with a new layout.
	 *
	 *Because this may happen unexpectedly applications, depending upon their logical structure,
	 *may need code in the Pause and Resume Subs to deal with the fact that modal dialog closure
	 *may not always be detected. Setting a process global when a modal dialog is shown and
	 *clearing it when it returns with some checking code in the Resume Sub is one way of dealing
	 *with this possibility.
	 *
	 *The above discussion also applies the Basic4android modal dialogs InputList, InputMultiList,
	 *Msgbox and Msgbox2.
	 */
	public void LIBRARY_DOC()
	{
	}

	// Private fields

	private String input = "";
	private int response;
	private boolean password = false;
	private int inputtype = 1;
	private String hint = "";
	private int hintcolor = Colors.DarkGray;

	public static final int INPUT_TYPE_NONE = 0;
	public static final int INPUT_TYPE_NUMBERS = 2;
	public static final int INPUT_TYPE_DECIMAL_NUMBERS = 12290;
	public static final int INPUT_TYPE_TEXT = 1;
	public static final int INPUT_TYPE_PHONE = 3;

	/**
	 *Shows a modal text input dialog with the specified message and title.
	 *Message - The dialog message.
	 *Title - The dialog title.
	 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
	 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
	 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
	 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
	 *Returns one of the DialogResponse values.
	 */
	public int Show(String message, String title, String Positive, String Cancel, String Negative, final BA ba, Bitmap icon)
	{
		AlertDialog ad = new AlertDialog.Builder(ba.context).create();

		float diptopx = ba.context.getResources().getDisplayMetrics().density;
		int dps20 = (int) (20 * diptopx);
		int dps10 = (int) (10 * diptopx);
		LinearLayout view = new LinearLayout(ba.context);
		LayoutParams lparams = new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT, ViewGroup.LayoutParams.FILL_PARENT);
		view.setLayoutParams(lparams);
		view.setOrientation(LinearLayout.VERTICAL);
		/*
		 * The margins version does not work if you are using the wrong
		 * layout params. You must always use the layout params of the
		 * layout you are adding a view to. In this case, you must use
		 * LinearLayout.LayoutParams, not ViewGroup.MarginLayoutParams. If
		 * you use the latter, an implicit conversion happens, which loses
		 * the margins.
		 */
		TextView text = new TextView(ba.context);
		LinearLayout.LayoutParams tparams = new LinearLayout.LayoutParams(ViewGroup.LayoutParams.WRAP_CONTENT,
				ViewGroup.LayoutParams.WRAP_CONTENT);
		tparams.setMargins(dps20, 0, dps10, 0); // must use correct params
		// type or margins are lost
		text.setLayoutParams(tparams);
		text.setGravity(Gravity.LEFT);
		text.setText(message);
		view.addView(text);
		
		EditText edit = new EditText(ba.context);
		LinearLayout.LayoutParams eparams = new LinearLayout.LayoutParams(LayoutParams.FILL_PARENT, LayoutParams.WRAP_CONTENT);
		eparams.setMargins(dps10, 0, dps10, 0); // must use correct params type or margins are lost
		edit.setLayoutParams(eparams);
		edit.setGravity(Gravity.FILL_HORIZONTAL);
		edit.setHorizontallyScrolling(true);
		edit.setText(input);
		edit.setInputType(inputtype);
		if (password)
			edit.setTransformationMethod(PasswordTransformationMethod.getInstance());
		else
			edit.setTransformationMethod(null);
		if (hint != "")
		{
			edit.setHint(hint);
			edit.setHintTextColor(hintcolor);
		}
		view.addView(edit);
		ad.setView(view);

		Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
		ad.setTitle(title);
		if (Positive.length() > 0)
			ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
		if (Negative.length() > 0)
			ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
		if (Cancel.length() > 0)
			ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
		if (icon != null)
		{
			BitmapDrawable bd = new BitmapDrawable(icon);
			ad.setIcon(bd);
		}
		Msgbox.msgbox(ad, false);
		input = edit.getText().toString();
		response = dr.res;
		// For some unknown reason closing the keyboard doesn't work without the Msgbox calls
		// Weirdly, because the dialog has just done this, if we do it again we can close the keyboard 
		Msgbox.sendCloseMyLoopMessage();
		Msgbox.waitForMessage(false, true);
		InputMethodManager imm = (InputMethodManager) BA.applicationContext.getSystemService("input_method");
		imm.hideSoftInputFromWindow(ba.vg.getWindowToken(), 0);
		return dr.res;
	}

	/**
	 *Returns the version of the library.
	 */
	public double getVersion()
	{
		return version;
	}

	/**
	 *Sets the initial text when the dialog is shown and returns the text entered by the user.
	 */
	public void setInput(String value)
	{
		input = value;
	}

	public String getInput()
	{
		return input;
	}

	/**
	 *Sets or returns the input type accepted by the input box.Possible values are:
	 *ThisDialogName.INPUT_TYPE_NUMBERS for integer numbers.
	 *ThisDialogName.INPUT_TYPE_DECIMAL_NUMBER for signed decimal numbers.
	 *ThisDialogName.INPUT_TYPE_TEXT for free text.
	 *ThisDialogName.INPUT_TYPE_PHONE for telephone numbers.
	 */
	public int getInputType()
	{
		return inputtype;
	}

	public void setInputType(int inputtype)
	{
		this.inputtype = inputtype;
	}

	/**
	 *Sets or returns whether this dialog hides the actual characters entered by the user.
	 */
	public boolean getPasswordMode()
	{
		return password;
	}

	public void setPasswordMode(boolean value)
	{
		this.password = value;
	}

	/**
	 *Returns the response code that the dialog returned when it last closed.
	 */
	public int getResponse()
	{
		return response;
	}

	/**
	 *Gets or sets the text that will appear when the dialog is empty.
	 */
	public String getHint()
	{
		return hint;
	}

	public void setHint(String hint)
	{
		this.hint = hint;
	}

	/**
	 *Gets or sets the hint text color.
	 */
	public int getHintColor()
	{
		return hintcolor;
	}

	public void setHintColor(int hintcolor)
	{
		this.hintcolor = hintcolor;
	}

	/**
	 *This modal dialog allows the collection of user entered data in the form of a date.
	 */
	@ShortName("DateDialog")
	@ActivityObject
	public static class DateDialog
	{
		private int year;
		private int month;
		private int day;
		private int response;

		/**
		 *Shows a modal date input dialog with the specified message and title.
		 *Message - The dialog message.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String Message, String Title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon)
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			float diptopx = ba.context.getResources().getDisplayMetrics().density;
			int dps20 = (int) (20 * diptopx);
			int dps10 = (int) (10 * diptopx);
			LinearLayout view = new LinearLayout(ba.context);
			LayoutParams lparams = new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT, ViewGroup.LayoutParams.WRAP_CONTENT);
			view.setLayoutParams(lparams);
			view.setOrientation(LinearLayout.VERTICAL);
			/*
			 * The margins version does not work if you are using the wrong
			 * layout params. You must always use the layout params of the
			 * layout you are adding a view to. In this case, you must use
			 * LinearLayout.LayoutParams, not ViewGroup.MarginLayoutParams. If
			 * you use the latter, an implicit conversion happens, which loses
			 * the margins.
			 */
			TextView text = new TextView(ba.context);
			LayoutParams tparams = new LayoutParams(ViewGroup.LayoutParams.WRAP_CONTENT, ViewGroup.LayoutParams.WRAP_CONTENT);
			tparams.setMargins(dps20, 0, dps10, 0); // must use correct params
			// type or margins are lost
			text.setLayoutParams(tparams);
			text.setGravity(Gravity.LEFT);
			text.setText(Message);
			view.addView(text);

			DatePicker dp = new DatePicker(ba.context);
			dp.updateDate(year, month, day);
			view.addView(dp);
			ad.setView(view);

			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(Title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			dp.clearFocus();
			year = dp.getYear();
			month = dp.getMonth();
			day = dp.getDayOfMonth();
			response = dr.res;
			return dr.res;
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Sets the year value of the dialog when is intially shown.
		 *Returns the year value of the dialog when it is closed.
		 */
		public int getYear()
		{
			return year;
		}

		public void setYear(int year)
		{
			this.year = year;
		}

		/**
		 *Sets the month value of the dialog when is intially shown.
		 *Returns the month value of the dialog when it is closed.
		 */
		public int getMonth()
		{
			return month + 1;
		}

		public void setMonth(int month)
		{
			this.month = month - 1;
		}

		/**
		 *Sets the day of month value of the dialog when is intially shown.
		 *Returns the day of month value of the dialog when it is closed.
		 */
		public int getDayOfMonth()
		{
			return day;
		}

		public void setDayOfMonth(int day)
		{
			this.day = day;
		}

		/**
		 *Sets the date values of the dialog when is intially shown.
		 */
		public void SetDate(int dayofmonth, int month, int year)
		{
			this.day = dayofmonth;
			this.month = month - 1;
			this.year = year;
		}

		/**
		 *Sets the date value of the dialog when is intially shown.
		 *Returns the date value in ticks of the dialog when it is closed.
		 */
		public long getDateTicks()
		{
			return new GregorianCalendar(year, month, day, 0, 0).getTimeInMillis();
		}

		public void setDateTicks(long ticks)
		{
			this.day = DateTime.GetDayOfMonth(ticks);
			this.month = DateTime.GetMonth(ticks) - 1;
			this.year = DateTime.GetYear(ticks);
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

	}

	/**
	 *This modal dialog allows the collection of user entered data in the form of a time.
	 *The time may be entered in 12 or 24 hour format as determined by the programmer.
	 */
	@ShortName("TimeDialog")
	@ActivityObject
	public static class TimeDialog
	{
		private int hour;
		private int minute;
		private boolean hours24;
		private int response;

		/**
		 *Shows a modal time input dialog with the specified message and title.
		 *Message - The dialog message.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String Message, String Title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon)
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			float diptopx = ba.context.getResources().getDisplayMetrics().density;
			int dps20 = (int) (20 * diptopx);
			int dps10 = (int) (10 * diptopx);
			LinearLayout view = new LinearLayout(ba.context);
			LayoutParams lparams = new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT, ViewGroup.LayoutParams.WRAP_CONTENT);
			view.setLayoutParams(lparams);
			view.setOrientation(LinearLayout.VERTICAL);
			/*
			 * The margins version does not work if you are using the wrong
			 * layout params. You must always use the layout params of the
			 * layout you are adding a view to. In this case, you must use
			 * LinearLayout.LayoutParams, not ViewGroup.MarginLayoutParams. If
			 * you use the latter, an implicit conversion happens, which loses
			 * the margins.
			 */
			TextView text = new TextView(ba.context);
			LayoutParams tparams = new LayoutParams(ViewGroup.LayoutParams.WRAP_CONTENT, ViewGroup.LayoutParams.WRAP_CONTENT);
			tparams.setMargins(dps20, 0, dps10, 0); // must use correct params
			// type or margins are lost
			text.setLayoutParams(tparams);
			text.setGravity(Gravity.LEFT);
			text.setText(Message);
			view.addView(text);

			TimePicker tp = new TimePicker(ba.context);
			tp.setIs24HourView(hours24);
			tp.setCurrentHour(hour);
			tp.setCurrentMinute(minute);
			view.addView(tp);
			ad.setView(view);

			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(Title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			tp.clearFocus();
			hour = tp.getCurrentHour();
			minute = tp.getCurrentMinute();
			response = dr.res;
			return dr.res;
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Sets the hour value of the dialog when is intially shown.
		 *Returns the hour value of the dialog when it is closed.
		 */
		public int getHour()
		{
			return hour;
		}

		public void setHour(int hour)
		{
			this.hour = hour;
		}

		/**
		 *Sets the minute value of the dialog when is intially shown.
		 *Returns the minute value of the dialog when it is closed.
		 */
		public int getMinute()
		{
			return minute;
		}

		public void setMinute(int minute)
		{
			this.minute = minute;
		}

		/**
		 *Sets or returns whether the dialog shows the time in 24 hour format.
		 */
		public boolean getIs24Hours()
		{
			return hours24;
		}

		public void setIs24Hours(boolean is24hours)
		{
			this.hours24 = is24hours;
		}

		/**
		 *Sets the time value of the dialog when is intially shown.
		 *Returns the time value in ticks of the dialog when it is closed.
		 */
		public long getTimeTicks()
		{
			return new GregorianCalendar(0, 0, 0, hour, minute).getTimeInMillis();
		}

		public void setTimeTicks(long ticks)
		{
			this.hour = DateTime.GetHour(ticks);
			this.minute = DateTime.GetMinute(ticks);
		}

		/**
		 *Sets the time values of the dialog when is intially shown.
		 */
		public void SetTime(int hour, int minutes, boolean hours24)
		{
			this.hour = hour;
			this.minute = minutes;
			this.hours24 = hours24;
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

	}

	/**
	 *This modal dialog allows the user to define a colour by its Red, Green and Blue components.
	 */
	@ShortName("ColorDialog")
	@ActivityObject
	public static class ColorDialog
	{
		// Private fields	

		private int response;
		private int red;
		private int green;
		private int blue;

		private GradientDrawable getColor(int color, float radius)
		{
			GradientDrawable gd = new GradientDrawable(GradientDrawable.Orientation.BL_TR, new int[] { color, color });
			gd.setCornerRadius(radius);
			return gd;
		}

		/**
		 *Shows a modal color dialog with the specified title.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon)
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			float diptopx = ba.context.getResources().getDisplayMetrics().density;
			int dps20 = (int) (20 * diptopx);
			int dps10 = (int) (10 * diptopx);
			final int dps05 = (int) (5 * diptopx);
			LinearLayout view = new LinearLayout(ba.context);
			LayoutParams lparams = new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT, ViewGroup.LayoutParams.WRAP_CONTENT);
			view.setLayoutParams(lparams);
			view.setOrientation(LinearLayout.VERTICAL);
			/*
			 * The margins version does not work if you are using the wrong
			 * layout params. You must always use the layout params of the
			 * layout you are adding a view to. In this case, you must use
			 * LinearLayout.LayoutParams, not ViewGroup.MarginLayoutParams. If
			 * you use the latter, an implicit conversion happens, which loses
			 * the margins.
			 */
			final TextView ctext = new TextView(ba.context);
			LayoutParams cparams = new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT, ViewGroup.LayoutParams.FILL_PARENT);
			cparams.setMargins(dps20, dps10, dps20, dps20); // must use correct params type or margins are lost
			cparams.height = dps20 * 2;
			ctext.setLayoutParams(cparams);
			//ctext.setGravity(Gravity.CENTER);
			ctext.setBackgroundDrawable(getColor(Color.argb(255, red, green, blue), dps05));
			view.addView(ctext);

			final TextView rtext = new TextView(ba.context);
			LayoutParams tparams = new LayoutParams(ViewGroup.LayoutParams.WRAP_CONTENT, ViewGroup.LayoutParams.WRAP_CONTENT);
			tparams.setMargins(dps20, 0, dps20, 0); // must use correct params type or margins are lost
			rtext.setLayoutParams(tparams);
			rtext.setGravity(Gravity.LEFT);
			rtext.setText(Integer.toString(red));
			view.addView(rtext);

			SeekBar sred = new SeekBar(ba.context);
			sred.setMax(255);
			sred.setProgress(red);
			sred.setBackgroundDrawable(getColor(Colors.Red, dps05));
			LayoutParams rparams = new LayoutParams(LayoutParams.FILL_PARENT, LayoutParams.WRAP_CONTENT);
			rparams.setMargins(dps10, dps10, dps10, dps10); // must use correct params type or margins are lost
			sred.setLayoutParams(rparams);
			view.addView(sred);
			sred.setOnSeekBarChangeListener(new OnSeekBarChangeListener()
			{
				@Override
				public void onStopTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onStartTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser)
				{
					red = progress;
					rtext.setText(Integer.toString(progress));
					ctext.setBackgroundDrawable(getColor(Color.argb(255, red, green, blue), dps05));
				}
			});

			final TextView gtext = new TextView(ba.context);
			gtext.setLayoutParams(tparams);
			gtext.setGravity(Gravity.LEFT);
			gtext.setText(Integer.toString(green));
			view.addView(gtext);
			SeekBar sgreen = new SeekBar(ba.context);
			sgreen.setMax(255);
			sgreen.setProgress(green);
			sgreen.setBackgroundDrawable(getColor(Colors.Green, dps05));
			//LayoutParams eparams = new LayoutParams(LayoutParams.FILL_PARENT,
			//		LayoutParams.WRAP_CONTENT);
			//rparams.setMargins(dps10, 0, dps10, 0); // must use correct params type or margins are lost
			sgreen.setLayoutParams(rparams);
			view.addView(sgreen);
			sgreen.setOnSeekBarChangeListener(new OnSeekBarChangeListener()
			{
				@Override
				public void onStopTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onStartTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser)
				{
					green = progress;
					gtext.setText(Integer.toString(progress));
					ctext.setBackgroundDrawable(getColor(Color.argb(255, red, green, blue), dps05));
				}
			});

			final TextView btext = new TextView(ba.context);
			btext.setLayoutParams(tparams);
			btext.setGravity(Gravity.LEFT);
			btext.setText(Integer.toString(blue));
			view.addView(btext);
			SeekBar sblue = new SeekBar(ba.context);
			sblue.setMax(255);
			sblue.setProgress(blue);
			sblue.setBackgroundDrawable(getColor(Colors.Blue, dps05));
			//LayoutParams eparams = new LayoutParams(LayoutParams.FILL_PARENT,
			//		LayoutParams.WRAP_CONTENT);
			//rparams.setMargins(dps10, 0, dps10, 0); // must use correct params
			// type or margins are lost
			sblue.setLayoutParams(rparams);
			sblue.setOnSeekBarChangeListener(new OnSeekBarChangeListener()
			{
				@Override
				public void onStopTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onStartTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser)
				{
					blue = progress;
					btext.setText(Integer.toString(progress));
					ctext.setBackgroundDrawable(getColor(Color.argb(255, red, green, blue), dps05));
				}
			});
			view.addView(sblue);

			ad.setView(view);

			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			red = sred.getProgress();
			green = sgreen.getProgress();
			blue = sblue.getProgress();
			response = dr.res;
			return dr.res;
		}

		/**
		 *Sets the value of the red component of the dialog when is intially shown.
		 *Returns the value of the red component of the dialog when it was closed.
		 */
		public int getRed()
		{
			return red;
		}

		public void setRed(int red)
		{
			this.red = red;
		}

		/**
		 *Sets the value of the green component of the dialog when is intially shown.
		 *Returns the value of the green component of the dialog when it was closed.
		 */
		public int getGreen()
		{
			return green;
		}

		public void setGreen(int green)
		{
			this.green = green;
		}

		/**
		 *Sets the value of the blue component of the dialog when is intially shown.
		 *Returns the value of the blue component of the dialog when it was closed.
		 */
		public int getBlue()
		{
			return blue;
		}

		public void setBlue(int blue)
		{
			this.blue = blue;
		}

		/**
		 *Sets the value of the red, green and blue components of the dialog when is intially shown.
		 *Returns the color of the red, green and blue components of the dialog when it was closed.
		 *Alpha of the provided color is ignored on set and implicitly set to 255 (opaque) on get.
		 */
		public int getRGB()
		{
			return Color.rgb(red, green, blue);
		}

		public void setRGB(int color)
		{
			this.red = Color.red(color);
			this.green = Color.green(color);
			this.blue = Color.blue(color);
		}

		/**
		 *Returns an integer value representing the color built from the three components and with the specified alpha value.
		 *Alpha - A value between 0 to 255 where 0 is fully transparent and 255 is fully opaque.
		 */
		public int ARGB(int alpha)
		{
			return Color.argb(alpha, red, green, blue);
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

	}

	/**
	 *This modal dialog allows the user to define a colour by its Hue, Saturation and Value components.
	 */
	@ShortName("ColorDialogHSV")
	@ActivityObject
	public static class ColorDialogHSV
	{
		// Private fields	

		private int response;
		private float[] hsv = new float[3];

		private GradientDrawable getColor(int color, float radius)
		{
			GradientDrawable gd = new GradientDrawable(GradientDrawable.Orientation.BL_TR, new int[] { color, color });
			gd.setCornerRadius(radius);
			return gd;
		}

		private GradientDrawable getGradient(int[] colors, float radius)
		{
			GradientDrawable gd = new GradientDrawable(GradientDrawable.Orientation.LEFT_RIGHT, colors);
			gd.setCornerRadius(radius);
			return gd;
		}

		/**
		 *Shows a modal color dialog with the specified title.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon)
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			float diptopx = ba.context.getResources().getDisplayMetrics().density;
			int dps20 = (int) (20 * diptopx);
			int dps10 = (int) (10 * diptopx);
			final int dps05 = (int) (5 * diptopx);
			LinearLayout view = new LinearLayout(ba.context);
			LayoutParams lparams = new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT, ViewGroup.LayoutParams.WRAP_CONTENT);
			view.setLayoutParams(lparams);
			view.setOrientation(LinearLayout.VERTICAL);
			/*
			 * The margins version does not work if you are using the wrong
			 * layout params. You must always use the layout params of the
			 * layout you are adding a view to. In this case, you must use
			 * LinearLayout.LayoutParams, not ViewGroup.MarginLayoutParams. If
			 * you use the latter, an implicit conversion happens, which loses
			 * the margins.
			 */
			final TextView ctext = new TextView(ba.context);
			LayoutParams cparams = new LayoutParams(ViewGroup.LayoutParams.FILL_PARENT, ViewGroup.LayoutParams.FILL_PARENT);
			cparams.setMargins(dps20, dps10, dps20, dps20); // must use correct params type or margins are lost
			cparams.height = dps20 * 2;
			ctext.setLayoutParams(cparams);
			//ctext.setGravity(Gravity.CENTER);
			ctext.setBackgroundDrawable(getColor(Color.HSVToColor(hsv), dps05));
			view.addView(ctext);

			final TextView rtext = new TextView(ba.context);
			LayoutParams tparams = new LayoutParams(ViewGroup.LayoutParams.WRAP_CONTENT, ViewGroup.LayoutParams.WRAP_CONTENT);
			tparams.setMargins(dps20, 0, dps20, 0); // must use correct params type or margins are lost
			rtext.setLayoutParams(tparams);
			rtext.setGravity(Gravity.LEFT);
			rtext.setText("Hue = " + Float.toString(hsv[0]));
			view.addView(rtext);

			SeekBar sred = new SeekBar(ba.context);
			sred.setMax(3600);
			sred.setProgress((int) (hsv[0] * 10f));
			//sred.setBackgroundDrawable(getColor(Colors.Red, dps05));
			sred.setBackgroundDrawable(getGradient(new int[] { Colors.Red, Colors.Green, Colors.Blue, Colors.Red }, dps05));
			LayoutParams rparams = new LayoutParams(LayoutParams.FILL_PARENT, LayoutParams.WRAP_CONTENT);
			rparams.setMargins(dps10, dps10, dps10, dps10); // must use correct params type or margins are lost
			sred.setLayoutParams(rparams);
			view.addView(sred);
			sred.setOnSeekBarChangeListener(new OnSeekBarChangeListener()
			{
				@Override
				public void onStopTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onStartTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser)
				{
					hsv[0] = progress / 10f;
					rtext.setText("Hue = " + Float.toString(hsv[0]));
					ctext.setBackgroundDrawable(getColor(Color.HSVToColor(hsv), dps05));
				}
			});

			final TextView gtext = new TextView(ba.context);
			gtext.setLayoutParams(tparams);
			gtext.setGravity(Gravity.LEFT);
			gtext.setText("Saturation = " + Float.toString(hsv[1]));
			view.addView(gtext);
			SeekBar sgreen = new SeekBar(ba.context);
			sgreen.setMax(1000);
			sgreen.setProgress((int) (hsv[1] * 1000f));
			sgreen.setBackgroundDrawable(getGradient(new int[] { Colors.DarkGray, Colors.White }, dps05));
			//LayoutParams eparams = new LayoutParams(LayoutParams.FILL_PARENT,
			//		LayoutParams.WRAP_CONTENT);
			//rparams.setMargins(dps10, 0, dps10, 0); // must use correct params type or margins are lost
			sgreen.setLayoutParams(rparams);
			view.addView(sgreen);
			sgreen.setOnSeekBarChangeListener(new OnSeekBarChangeListener()
			{
				@Override
				public void onStopTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onStartTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser)
				{
					hsv[1] = progress / 1000f;
					gtext.setText("Saturation = " + Float.toString(hsv[1]));
					ctext.setBackgroundDrawable(getColor(Color.HSVToColor(hsv), dps05));
				}
			});

			final TextView btext = new TextView(ba.context);
			btext.setLayoutParams(tparams);
			btext.setGravity(Gravity.LEFT);
			btext.setText("Value = " + Float.toString(hsv[2]));
			view.addView(btext);
			SeekBar sblue = new SeekBar(ba.context);
			sblue.setMax(1000);
			sblue.setProgress((int) (hsv[2] * 1000f));
			sblue.setBackgroundDrawable(getGradient(new int[] { Colors.DarkGray, Colors.White }, dps05));
			//LayoutParams eparams = new LayoutParams(LayoutParams.FILL_PARENT,
			//		LayoutParams.WRAP_CONTENT);
			//rparams.setMargins(dps10, 0, dps10, 0); // must use correct params
			// type or margins are lost
			sblue.setLayoutParams(rparams);
			sblue.setOnSeekBarChangeListener(new OnSeekBarChangeListener()
			{
				@Override
				public void onStopTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onStartTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser)
				{
					hsv[2] = progress / 1000f;
					btext.setText("Value = " + Float.toString(hsv[2]));
					ctext.setBackgroundDrawable(getColor(Color.HSVToColor(hsv), dps05));
				}
			});
			view.addView(sblue);

			ad.setView(view);

			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			response = dr.res;
			return dr.res;
		}

		/**
		 *Sets the value of the hue component of the dialog when is intially shown.
		 *Returns the value of the hue component of the dialog when it was closed.
		 *The range of valid numbers for hue is 0.0 to 360.0.
		 */
		public float getHue()
		{
			return hsv[0];
		}

		public void setHue(float hue)
		{
			this.hsv[0] = hue;
		}

		/**
		 *Sets the value of the saturation component of the dialog when is intially shown.
		 *Returns the value of the saturation component of the dialog when it was closed.
		 *The range of valid numbers for saturation is 0.0 to 1.0.
		 */
		public float getSaturation()
		{
			return hsv[1];
		}

		public void setSaturation(float saturation)
		{
			this.hsv[1] = saturation;
		}

		/**
		 *Sets the value of the value component of the dialog when is intially shown.
		 *Returns the value of the value component of the dialog when it was closed.
		 *The range of valid numbers for value is 0.0 to 1.0.
		 */
		public float getValue()
		{
			return hsv[2];
		}

		public void setValue(float value)
		{
			this.hsv[2] = value;
		}

		/**
		 *Sets the value of the red, green and blue components of the dialog when is intially shown.
		 *Returns the color of the red, green and blue components of the dialog when it was closed.
		 *Alpha of the provided color is ignored on set and implicitly set to 255 (opaque) on get.
		 */
		public int getRGB()
		{
			return Color.HSVToColor(hsv);
		}

		public void setRGB(int color)
		{
			Color.colorToHSV(color, hsv);
		}

		/**
		 *Returns an integer value representing the color built from the three components and with the specified alpha value.
		 *Alpha - A value between 0 to 255 where 0 is fully transparent and 255 is fully opaque.
		 */
		public int ARGB(int alpha)
		{
			return Color.HSVToColor(alpha, hsv);
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

	}

	/**
	 *This modal dialog allows the user to select a colour from a palette of colours.
	 *The color may be from a standard palette in the dialog or a custom programmed palette.
	 */
	@ShortName("ColorPickerDialog")
	@ActivityObject
	public static class ColorPickerDialog
	{
		// Private fields	

		private int response;
		private int chosen;
		int orange = Colors.RGB(255, 140, 0);
		int green = Colors.RGB(0, 128, 0);
		int springgreen = Colors.RGB(0, 255, 127);
		int maroon = Colors.RGB(128, 00, 0);
		int navy = Colors.RGB(0, 0, 128);
		int royalblue = Colors.RGB(65, 105, 225);

		private int[] stdpalette = { Colors.Red, Colors.Magenta, orange, maroon, Colors.Black, Colors.Green, Colors.Yellow,
				springgreen, green, Colors.LightGray, Colors.Blue, Colors.Cyan, royalblue, navy, Colors.White };
		private int[] palette;

		public ColorPickerDialog()
		{
			palette = new int[stdpalette.length];
			ResetPalette();
		}

		private GradientDrawable getColor(int color, float radius)
		{
			GradientDrawable gd = new GradientDrawable(GradientDrawable.Orientation.BL_TR, new int[] { color, color });
			gd.setCornerRadius(radius);
			return gd;
		}

		/**
		 *Shows a modal color picker dialog with the specified title.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon)
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			float diptopx = ba.context.getResources().getDisplayMetrics().density;
			int dps40 = (int) (40 * diptopx);
			int dps20 = (int) (20 * diptopx);
			int dps10 = (int) (10 * diptopx);
			final int dps05 = (int) (5 * diptopx);

			LinearLayout lview = new LinearLayout(ba.context);
			lview.setOrientation(LinearLayout.VERTICAL);
			final TextView ctext = new TextView(ba.context);
			TableRow.LayoutParams cparams = new TableRow.LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
					ViewGroup.LayoutParams.FILL_PARENT);
			cparams.setMargins(dps20, dps10, dps20, dps20); // must use correct params type or margins are lost
			cparams.height = dps40;
			cparams.width = -1;
			ctext.setLayoutParams(cparams);
			ctext.setBackgroundDrawable(getColor(chosen, dps05));
			lview.addView(ctext);

			TableLayout view = new TableLayout(ba.context);
			TableLayout.LayoutParams lparams = new TableLayout.LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
					ViewGroup.LayoutParams.WRAP_CONTENT);
			view.setLayoutParams(lparams);
			view.setOrientation(LinearLayout.VERTICAL);
			view.setStretchAllColumns(true);

			for (int j = 0; j < 3; j++)
			{
				TableRow row1 = new TableRow(ba.context);
				for (int i = 0; i < 5; i++)
				{
					final TextView btn0 = new TextView(ba.context);
					TableRow.LayoutParams bparams = new TableRow.LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
							ViewGroup.LayoutParams.FILL_PARENT);
					bparams.setMargins(dps05, dps05, dps05, 05); // must use correct params type or margins are lost
					bparams.height = dps40;
					bparams.width = LayoutParams.FILL_PARENT;
					btn0.setLayoutParams(bparams);
					btn0.setTag(palette[j * 5 + i]);
					btn0.setBackgroundDrawable(getColor(palette[j * 5 + i], dps05));
					btn0.setOnClickListener(new OnClickListener()
					{
						@Override
						public void onClick(View v)
						{
							chosen = (Integer) v.getTag();
							ctext.setBackgroundDrawable(getColor(chosen, dps05));
						}
					});
					row1.addView(btn0);
				}
				view.addView(row1);
			}
			lview.addView(view);
			ad.setView(lview);

			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			response = dr.res;
			return dr.res;
		}

		/**
		 *Copies the colours in the array provided to the palette of colors in the dialog.
		 *The provided array should contain 15 colors.
		 *Returns an integer array that is a copy of the present palette.
		 */
		public int[] getPalette()
		{
			int[] newpalette = new int[palette.length];
			for (int i = 0; i < palette.length; i++)
				newpalette[i] = palette[i];
			return newpalette;
		}

		public void setPalette(int[] palette)
		{
			int numcols = Math.min(15, palette.length);
			for (int i = 0; i < numcols; i++)
				this.palette[i] = palette[i];
		}

		/**
		 *Reset the palette of colors to the standard palette of the dialog.
		 */
		public void ResetPalette()
		{
			for (int i = 0; i < stdpalette.length; i++)
				palette[i] = stdpalette[i];
		}

		/**
		 *Sets the value of the chosen color of the dialog when is intially shown.
		 *Returns the value of the chosen color of the dialog when it was closed.
		 */
		public int getRGB()
		{
			return chosen;
		}

		public void setRGB(int color)
		{
			this.chosen = color | 0xff000000;
		}

		/**
		 *Gets the value of the color at the specified index in the current palette.
		 */
		public int GetPaletteAt(int index)
		{
			return palette[index];
		}

		/**
		 *Sets the value of the color at the specified index in the current palette.
		 *This allows replacing just one or two colors without defining an entire palette.
		 */
		public void SetPaletteAt(int index, int color)
		{
			palette[index] = color | 0xff000000;
		}

		/**
		 *Returns an integer value representing the color built from the chosen color and with the specified alpha value.
		 *Alpha - A value between 0 to 255 where 0 is fully transparent and 255 is fully opaque.
		 */
		public int ARGB(int alpha)
		{
			int red = chosen >> 16 & 0xff;
			int green = chosen >> 8 & 0xff;
			int blue = chosen & 0xff;
			return Color.argb(alpha, red, green, blue);
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

	}

	/**
	 *This configurable modal dialog allows the user to enter a number.
	 *The dialog is configurable to show any number of digits  between a minimum of one and a maximum of eight.
	 *The display of a decimal point is optional and the character displayed as the decimal indicator is configurable.
	 *Note that the number accepted and returned by the dialog is an integer value and so may need scaling appropriately.
	 */
	@ShortName("NumberDialog")
	@ActivityObject
	public static class NumberDialog
	{
		// Private fields	

		private int response;
		private int numdigits = 5;
		final private int maxdigits = 9;
		private int decimal = 0;
		private char decimalchar = '.';
		private boolean showsign = false;
		private boolean negative = false;
		private int[] numbers = new int[maxdigits];
		private TextView[] figures = new TextView[maxdigits];

		/**
		 *Shows a modal number picker dialog with the specified title.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon)
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			float diptopx = ba.context.getResources().getDisplayMetrics().density;
			int dps50 = (int) (50 * diptopx);
			int dps40 = (int) (40 * diptopx);
			final int dps05 = (int) (5 * diptopx);

			TableLayout view = new TableLayout(ba.context);
			TableLayout.LayoutParams lparams = new TableLayout.LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
					ViewGroup.LayoutParams.WRAP_CONTENT);
			view.setLayoutParams(lparams);
			view.setOrientation(LinearLayout.VERTICAL);
			view.setStretchAllColumns(true);

			// + buttons
			TableRow row1 = new TableRow(ba.context);
			for (int i = numdigits - 1; i >= 0; i--)
			{
				final Button btn0 = new Button(ba.context);
				TableRow.LayoutParams bparams = new TableRow.LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
						ViewGroup.LayoutParams.FILL_PARENT);
				bparams.setMargins(0, 0, 0, 0); // must use correct params type or margins are lost
				bparams.height = dps50;
				bparams.width = LayoutParams.FILL_PARENT;
				btn0.setLayoutParams(bparams);
				btn0.setPadding(0, -dps05, 0, 0);
				btn0.setText("+");
				//btn0.setTextColor(Colors.DarkGray);
				btn0.setTextSize(32);
				btn0.setTag(i);
				btn0.setOnClickListener(new OnClickListener()
				{
					@Override
					public void onClick(View v)
					{
						int i = (Integer) v.getTag();
						if (showsign && i == numdigits - 1)
						{
							figures[i].setText("+");
						} else
						{
							numbers[i] += 1;
							if (numbers[i] > 9)
								numbers[i] = 0;
							figures[i].setText(Integer.toString(numbers[i]));
							if (decimal > 0 && decimal == i)
								figures[i].setText(Integer.toString(numbers[i]) + decimalchar);
						}
					}
				});
				row1.addView(btn0);
			}
			view.addView(row1);

			// numbers
			row1 = new TableRow(ba.context);
			for (int i = numdigits - 1; i >= 0; i--)
			{
				final TextView btn0 = new TextView(ba.context);
				TableRow.LayoutParams bparams = new TableRow.LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
						ViewGroup.LayoutParams.FILL_PARENT);
				bparams.setMargins(0, 0, 0, 0); // must use correct params type or margins are lost
				bparams.height = dps40;
				bparams.width = LayoutParams.FILL_PARENT;
				btn0.setLayoutParams(bparams);
				btn0.setPadding(0, 0, 0, 0);
				btn0.setGravity(Gravity.CENTER);
				//btn0.setBackgroundColor(Colors.LightGray);
				//btn0.setTextColor(Colors.DarkGray);
				btn0.setText(Integer.toString(numbers[i]));
				if (decimal > 0 && decimal == i)
					btn0.setText(Integer.toString(numbers[i]) + decimalchar);
				btn0.setTextSize(32);
				btn0.setTag(i);
				row1.addView(btn0);
				figures[i] = btn0;
				if (showsign && i == numdigits - 1)
				{
					figures[i].setText(negative ? "-" : "+");
				}

			}
			view.addView(row1);

			// - buttons
			row1 = new TableRow(ba.context);
			for (int i = numdigits - 1; i >= 0; i--)
			{
				final Button btn0 = new Button(ba.context);
				TableRow.LayoutParams bparams = new TableRow.LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
						ViewGroup.LayoutParams.FILL_PARENT);
				bparams.setMargins(0, 0, 0, 0); // must use correct params type or margins are lost
				bparams.height = dps50;
				bparams.width = LayoutParams.FILL_PARENT;
				btn0.setLayoutParams(bparams);
				btn0.setPadding(0, -dps05, 0, 0);
				btn0.setText("-");
				btn0.setTextSize(32);
				btn0.setTag(i);
				btn0.setOnClickListener(new OnClickListener()
				{
					@Override
					public void onClick(View v)
					{
						int i = (Integer) v.getTag();
						if (showsign && i == numdigits - 1)
						{
							figures[i].setText("-");
						} else
						{
							numbers[i] -= 1;
							if (numbers[i] < 0)
								numbers[i] = 9;
							figures[i].setText(Integer.toString(numbers[i]));
							if (decimal > 0 && decimal == i)
								figures[i].setText(Integer.toString(numbers[i]) + decimalchar);

						}
					}
				});
				row1.addView(btn0);
			}
			view.addView(row1);
			ad.setView(view);

			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			response = dr.res;
			return dr.res;
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Sets the number initially displayed in the dialog when it is shown.
		 *If the number is negative and ShowSign is False then the absolute value is displayed
		 *Gets the number entered by the user after the dialog is closed.
		 *If ShowSign is True the sign of the number corresponds to the sign entered by the user.
		 */
		public int getNumber()
		{
			int num = 0;
			int count = numdigits - 1;
			count -= showsign ? 1 : 0;
			for (int i = count; i >= 0; i--)
				num = num * 10 + numbers[i];
			if (showsign && figures[numdigits - 1].getText() == "-")
				return -num;
			return num;
		}

		public void setNumber(int number)
		{
			negative = number < 0 ? true : false;
			number = Math.abs(number);
			for (int i = 0; i < numdigits; i++)
			{
				numbers[i] = number % 10;
				number /= 10;
			}
		}

		/**
		 *Gets or sets the number of digits displayed in the dialog when it is open.
		 *One is the minimum, nine is the maximum. The default is five.
		 *If ShowSign is True then the leftmost digit will display a "+" or "-".
		 */
		public int getDigits()
		{
			return numdigits;
		}

		public void setDigits(int digits)
		{
			numdigits = Math.min(maxdigits, Math.max(digits, 1));
		}

		/**
		 *Gets or sets the position of a displayed decimal point in the dialog.
		 *Zero displays no decimals, one indicates a single decimal, and so on.
		 */
		public int getDecimal()
		{
			return decimal;
		}

		public void setDecimal(int digits)
		{
			decimal = Math.min(maxdigits - 1, Math.max(digits, 0));
		}

		/**
		 *Gets or sets the displayed decimal character in the dialog.
		 *The default is ".".
		 */
		public char getDecimalChar()
		{
			return decimalchar;
		}

		public void setDecimalChar(char decimalchar)
		{
			this.decimalchar = decimalchar;
		}

		/**
		 *Gets or sets whether the displayed number includes a sign character.
		 *The default is False, no absolute value of the input number is displayed.
		 */
		public boolean getShowSign()
		{
			return showsign;
		}

		public void setShowSign(boolean show)
		{
			this.showsign = show;
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

	}

	/**
	 *This modal dialog allows the user to choose a folder and choose or enter a filename.
	 */
	@ShortName("FileDialog")
	@ActivityObject
	public static class FileDialog
	{
		private int response;
		private String fileseparator = "↓ Files ↓";
		private String[] fileList;
		private File filepath = new File("//");
		private String chosenname = "";
		private String ftype = "";
		private String[] ftypes = ftype.split(",");
		private boolean fastscroll = false;
		private int scrollcol = Colors.ARGB(230, 72, 72, 72);
		private boolean onlyfolders = false;

		private class sorter implements Comparator<String>
		{
			@Override
			public int compare(String arg0, String arg1)
			{
				return arg0.compareToIgnoreCase(arg1);
			}
		}

		private void loadFileList(SimpleListAdapter adapter) throws RuntimeException
		{
			final ArrayList<String> aldirs = new ArrayList<String>();
			final ArrayList<String> alfiles = new ArrayList<String>();
			String[] dirs;
			String[] files;
			adapter.items.clear();
			if (filepath.exists())
			{
				FilenameFilter filter = new FilenameFilter()
				{
					public boolean accept(File dir, String filename)
					{
						File sel = new File(dir, filename);
						if (sel.isDirectory())
						{
							aldirs.add(filename);
							return true;
						}
						for (int i = 0; i < ftypes.length; i++)
						{
							if (filename.contains(ftypes[i]) || ftype == "")
							{
								alfiles.add(filename);
								return true;
							}
						}
						return false;
					}
				};
				fileList = filepath.list(filter);
				dirs = aldirs.toArray(new String[0]);
				files = alfiles.toArray(new String[0]);
				Arrays.sort(dirs, new sorter());
				Arrays.sort(files, new sorter());
				if (onlyfolders)
				{
					fileList = new String[dirs.length+ 1];
					fileList[0] = "..";
					System.arraycopy(dirs, 0, fileList, 1, dirs.length);
				} else
				{
					fileList = new String[dirs.length + files.length + 2];
					fileList[0] = "..";
					fileList[dirs.length + 1] = fileseparator;
					System.arraycopy(dirs, 0, fileList, 1, dirs.length);
					System.arraycopy(files, 0, fileList, dirs.length + 2, files.length);

				}

			} else
			{
				throw new RuntimeException("Path '" + filepath.getPath() + "'does not exist");
			}
			for (int i = 0; i < fileList.length; i++)
			{
				SimpleListAdapter.SingleLineData sl = new SimpleListAdapter.SingleLineData();
				sl.Text = fileList[i];
				sl.ReturnValue = null;
				adapter.items.add(sl);
			}
			adapter.notifyDataSetChanged();
		}

		/**
		 *Shows a modal file dialog with the specified title.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String Title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon) throws RuntimeException
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			float diptopx = ba.context.getResources().getDisplayMetrics().density;
			int dps10 = (int) (10 * diptopx);
			LinearLayout view = new LinearLayout(ba.context);
			ViewGroup.LayoutParams lparams = new ViewGroup.LayoutParams(ViewGroup.LayoutParams.FILL_PARENT,
					ViewGroup.LayoutParams.FILL_PARENT);
			view.setLayoutParams(lparams);
			view.setOrientation(LinearLayout.VERTICAL);

			final EditText edit = new EditText(ba.context);
			LinearLayout.LayoutParams eparams = new LinearLayout.LayoutParams(LayoutParams.FILL_PARENT, LayoutParams.WRAP_CONTENT);
			eparams.setMargins(dps10, 0, dps10, 0); // must use correct params type or margins are lost
			edit.setLayoutParams(eparams);
			edit.setGravity(Gravity.FILL_HORIZONTAL);
			edit.setHorizontallyScrolling(true);
			edit.setImeOptions(6);
			edit.setSingleLine();
			edit.setText(chosenname);
			view.addView(edit);

			final ListView list = new ListView(ba.context);
			list.setFastScrollEnabled(fastscroll);
			list.setCacheColorHint(scrollcol);
			final SimpleListAdapter adapter = new SimpleListAdapter(ba.context);
			list.setAdapter(adapter);
			LinearLayout.LayoutParams lvparams = new LinearLayout.LayoutParams(LayoutParams.FILL_PARENT, LayoutParams.WRAP_CONTENT);
			list.setLayoutParams(lvparams);
			loadFileList(adapter);
			list.setOnItemClickListener(new AdapterView.OnItemClickListener()
			{
				public void onItemClick(AdapterView<?> parent, View view, int position, long id)
				{
					String chosenfile = (String) adapter.getItem(position);
					if (chosenfile.equals(".."))
					{
						// we don;t need to do anything in this case
						// because the system recognizes ".." as a valid directory						
					}
					File sel = new File(filepath, chosenfile);
					if (sel.isDirectory())
					{
						edit.setText("");
						filepath = sel;
						try
						{
							loadFileList(adapter);
							list.setSelection(0); // list.scrollTo doesn't seem to work :(
						} catch (RuntimeException e)
						{
							if (e instanceof NoSuchElementException)
							{
								Common.Log(e.toString() + " " + filepath + " " + chosenfile);
							}
							throw e;
						}
					} else
					{
						if (!chosenfile.equals(fileseparator))
						{
							edit.setText(chosenfile);
							chosenname = chosenfile;
						}
					}
				}
			});

			view.addView(list);

			ad.setView(view);
			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(Title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			chosenname = edit.getText().toString();
			response = dr.res;
			return dr.res;
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Gets or sets the filter values of the dialog.
		 *The filter can be a single value ".txt"
		 *The filter can also be a comma separated list of values ".jpg,.png".
		 *Note that spaces in filter values are significant and are not ignored.
		 *If a filename contains the text of a filter value the file will be displayed.
		 *A value of an empty string, the default, will show all files.
		 */
		public String getFileFilter()
		{
			return ftype;
		}

		public void setFileFilter(String filetype)
		{
			ftype = filetype;
			ftypes = ftype.split(",");
		}

		/**
		 *Sets the file path of the dialog when it is intially shown.
		 *Returns the file path of the dialog when it is closed.
		 *Note that setting the file path also sets ChosenName to an empty string.
		 */
		public String getFilePath() throws Exception
		{
			return filepath.getCanonicalPath();
		}

		public void setFilePath(String path)
		{
			filepath = new File(path, "//");
			chosenname = "";
		}

		/**
		 *Gets or sets whether the fast scroll thumb is displayed by the dialog. 
		 */
		public boolean getFastScroll()
		{
			return fastscroll;
		}

		public void setFastScroll(boolean fastscroll)
		{
			this.fastscroll = fastscroll;
		}

		/**
		 *Sets the filename initially shown to the user.
		 *Returns the filename entered or chosen by the user.
		 */
		public String getChosenName()
		{
			return chosenname;
		}

		public void setChosenName(String filename)
		{
			chosenname = filename;
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

		/**
		 *Gets or sets the background color that will be used while scrolling the list.
		 *This is an optimization done to make the scrolling smoother.
		 *Set to Colors.Transparent if the background behind the list is not solid color.
		 *The default whatever is the default for the particular device
		 */
		public void setScrollingBackgroundColor(int scrollcolor)
		{
			scrollcol = scrollcolor;
		}

		public int getScrollingBackgroundColor()
		{
			return scrollcol;
		}

		/**
		 *Gets or sets whether to show only folders and not files in the dialog.
		 */
		public void setShowOnlyFolders(boolean onlyfolders)
		{
			this.onlyfolders = onlyfolders;
		}

		public boolean getShowOnlyFolders()
		{
			return onlyfolders;
		}

	}

	/**
	 *This modal dialog displays a custom set of controls laid out on a Basic4android Panel.
	 *The Panel is displayed at an abolute position and size within the dialog.
	 */
	@ShortName("CustomDialog")
	@ActivityObject
	public static class CustomDialog
	{
		private int response;
		private View cview;

		/**
		 *Shows a modal custom dialog with the specified title.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String Title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon) throws RuntimeException
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			AbsoluteLayout view = new AbsoluteLayout(ba.context);
			AbsoluteLayout.LayoutParams lparams = new AbsoluteLayout.LayoutParams(0, 0, 0, 0);
			view.setLayoutParams(lparams);

			view.addView(cview);

			ad.setView(view);
			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(Title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			view.removeView(cview);
			response = dr.res;
			// Closing keyboard doesn't work without the Msgbox calls
			Msgbox.sendCloseMyLoopMessage();
			Msgbox.waitForMessage(false, true);
			InputMethodManager imm = (InputMethodManager) BA.applicationContext.getSystemService("input_method");
			imm.hideSoftInputFromWindow(ba.vg.getWindowToken(), 0);
			return dr.res;
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Adds the custom layout view, most probably a Panel, to the custom dialog.
		 *Although named AddView to match Basic4androd syntax only one view can be added.
		 *Adding a view replaces any existing view previously added to the dialog.
		 */
		public void AddView(View view, int left, int top, int width, int height)
		{
			cview = view;
			AbsoluteLayout.LayoutParams cparams = new AbsoluteLayout.LayoutParams(width, height, left, top);
			cview.setLayoutParams(cparams);
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

	}

	/**
	 *This modal dialog displays a custom set of controls laid out on a Basic4android Panel.
	 *The Panel will be automatically centred in the displayed dialog.
	 */
	@ShortName("CustomDialog2")
	@ActivityObject
	public static class CustomDialog2
	{
		private int response;
		private View cview;

		/**
		 *Shows a modal custom dialog with the specified title.
		 *Title - The dialog title.
		 *Positive - The text to show for the "positive" button. Pass "" if you don't want to show the button.
		 *Cancel - The text to show for the "cancel" button. Pass "" if you don't want to show the button.
		 *Negative - The text to show for the "negative" button. Pass "" if you don't want to show the button.
		 *Icon - A bitmap that will be drawn near the title. Pass Null if you don't want to show an icon.
		 *Returns one of the DialogResponse values.
		 */
		public int Show(String Title, String Positive, String Cancel, String Negative, BA ba, Bitmap icon) throws RuntimeException
		{
			AlertDialog ad = new AlertDialog.Builder(ba.context).create();

			LinearLayout view = new LinearLayout(ba.context);
			ViewGroup.LayoutParams lparams = new ViewGroup.LayoutParams(LinearLayout.LayoutParams.FILL_PARENT,
					LinearLayout.LayoutParams.FILL_PARENT);
			view.setGravity(Gravity.CENTER);
			view.setLayoutParams(lparams);

			view.addView(cview);

			ad.setView(view);
			Msgbox.DialogResponse dr = new Msgbox.DialogResponse(false);
			ad.setTitle(Title);
			if (Positive.length() > 0)
				ad.setButton(DialogInterface.BUTTON_POSITIVE, Positive, dr);
			if (Negative.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEGATIVE, Negative, dr);
			if (Cancel.length() > 0)
				ad.setButton(DialogInterface.BUTTON_NEUTRAL, Cancel, dr);
			if (icon != null)
			{
				BitmapDrawable bd = new BitmapDrawable(icon);
				ad.setIcon(bd);
			}
			Msgbox.msgbox(ad, false);
			view.removeView(cview);
			response = dr.res;
			// Closing keyboard doesn't work without the Msgbox calls
			Msgbox.sendCloseMyLoopMessage();
			Msgbox.waitForMessage(false, true);
			InputMethodManager imm = (InputMethodManager) BA.applicationContext.getSystemService("input_method");
			imm.hideSoftInputFromWindow(ba.vg.getWindowToken(), 0);
			return dr.res;
		}

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Adds the custom layout view, most probably a Panel, to the custom dialog.
		 *Although named AddView to match Basic4androd syntax only one view can be added.
		 *Adding a view replaces any existing view previously added to the dialog.
		 */
		public void AddView(View view, int width, int height)
		{
			cview = view;
			LinearLayout.LayoutParams cparams = new LinearLayout.LayoutParams(width, height);
			cview.setLayoutParams(cparams);
		}

		/**
		 *Returns the response code that the dialog returned when it last closed.
		 */
		public int getResponse()
		{
			return response;
		}

	}

}
